module AST where

import Data.Char
import Text.Read
import Text.ParserCombinators.ReadP hiding ((+++), choice)

data AST = Function String [String] Expr
         | Eval Expr
  deriving Show
data Expr = Num Float
          | BinOp BinOpType Expr Expr
          | Var String
          | Call String [Expr]
  deriving Show
data BinOpType = Add | Sub | Mul
  deriving Show

instance Read AST where
  readPrec = parseFunction +++ (Eval <$> readPrec)
    where parseFunction = lift $ do
            skipSpaces
            string "def"
            skipSpaces
            name <- munch1 isAlpha
            params <- between (char '(') (char ')') $
                      sepBy (munch1 isAlpha) skipSpaces
            skipSpaces
            body <- between (char '{') (char '}') $
                      readS_to_P reads
            skipSpaces
            return (Function name params body)

instance Read Expr where
  readPrec = choice [ parseNum
                    , parseVar
                    , parseCall
                    , parseBinOp '+' Add
                    , parseBinOp '-' Sub
                    , parseBinOp '*' Mul
                    ]
    where parseNum = Num <$> readPrec
          parseVar = Var <$> lift (munch1 isAlpha)
          parseCall = do
            func <- lift (munch1 isAlpha)
            params <- lift $ between (char '(') (char ')') $
                        sepBy (readS_to_P reads)
                              (skipSpaces >> char ',' >> skipSpaces)
            return (Call func params)
          parseBinOp c typ = step $ do
            a <- prec 11 readPrec
            lift $ do
              skipSpaces
              char c
              skipSpaces
            b <- readPrec
            return (BinOp typ a b)
